@@ -34,6 +34,9 @@ import java.net.URLEncoder;
 import java.nio.charset.Charset;
 import java.nio.charset.CharsetEncoder;
 import java.text.DateFormat;
+import java.text.DecimalFormat;
+import java.text.DecimalFormatSymbols;
+import java.text.NumberFormat;
 import java.text.SimpleDateFormat;
 import java.util.ArrayList;
 import java.util.Arrays;
@@ -76,27 +79,47 @@ import com.eteks.sweethome3d.tools.URLContent;
 public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
   private static final String [] IGNORED_EXTENSIONS = {".max", ".lwo", ".dxf"};
   private static final Locale DEFAULT_LOCALE = new Locale("");
+  private static final NumberFormat DECIMAL_FORMAT = new DecimalFormat("0.#####", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
   
   private static final String ID          = "id"; 
   private static final String NAME        = "name"; 
   private static final String DESCRIPTION = "description"; 
   private static final String VERSION     = "version"; 
   private static final String LICENSE     = "license"; 
-  private static final String PROVIDER    = "provider"; 
+  private static final String PROVIDER    = "provider";
   
   /**
-   * Reads furniture library from the given file.
+   * Reads a furniture library from the given file, after clearing the given library.
    */
   public void readFurnitureLibrary(final FurnitureLibrary furnitureLibrary, 
-                                   String furnitureLibraryName,
+                                   String furnitureLibraryLocation,
                                    FurnitureLibraryUserPreferences preferences) throws RecorderException {
+    readFurnitureLibrary(furnitureLibrary, furnitureLibraryLocation, preferences, false);
+  }
+  
+  /**
+   * Merges a furniture library with one in the given file.
+   */
+  public void mergeFurnitureLibrary(FurnitureLibrary furnitureLibrary,
+                                    String furnitureLibraryLocation,
+                                    FurnitureLibraryUserPreferences preferences) throws RecorderException {
+    readFurnitureLibrary(furnitureLibrary, furnitureLibraryLocation, preferences, true);
+  }
+
+  /**
+   * Reads a furniture library from the given file.
+   */
+  private void readFurnitureLibrary(final FurnitureLibrary furnitureLibrary, 
+                                    String furnitureLibraryLocation, 
+                                    FurnitureLibraryUserPreferences preferences, 
+                                    final boolean mergeLibrary) throws RecorderException {
     try {
       // Retrieve furniture library with default reader and locale 
       Locale defaultLocale = Locale.getDefault();
       Locale.setDefault(DEFAULT_LOCALE);
       File furnitureLibraryFile = File.createTempFile("furniture", ".sh3f");
       furnitureLibraryFile.deleteOnExit();
-      copyFile(new File(furnitureLibraryName), furnitureLibraryFile);      
+      copyFile(new File(furnitureLibraryLocation), furnitureLibraryFile);
       URL furnitureLibraryUrl = furnitureLibraryFile.toURI().toURL();
       String furnitureResourcesLocalDirectory = preferences.getFurnitureResourcesLocalDirectory();
       URL furnitureResourcesUrlBase = furnitureResourcesLocalDirectory != null
@@ -109,7 +132,7 @@ public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
                                                                  int index,
                                                                  URL furnitureCatalogUrl,
                                                                  URL furnitureResourcesUrlBase) {
-            if (index == 1) {
+            if (index == 1 && !mergeLibrary) {
               furnitureLibrary.setId(getOptionalString(resource, ID));
               furnitureLibrary.setName(getOptionalString(resource, NAME));
               furnitureLibrary.setDescription(getOptionalString(resource, DESCRIPTION));
@@ -148,8 +171,10 @@ public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
       }
 
       // Replace furniture by the one read
-      for (CatalogPieceOfFurniture piece : furnitureLibrary.getFurniture()) {
-        furnitureLibrary.deletePieceOfFurniture(piece);
+      if (!mergeLibrary) {
+        for (CatalogPieceOfFurniture piece : furnitureLibrary.getFurniture()) {
+          furnitureLibrary.deletePieceOfFurniture(piece);
+        }
       }
       for (CatalogPieceOfFurniture piece : furniture) {
         furnitureLibrary.addPieceOfFurniture(piece);
@@ -189,9 +214,9 @@ public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
       
       Locale.setDefault(defaultLocale);
     } catch (IOException ex) {
-      throw new RecorderException("Invalid furniture library file " + furnitureLibraryName, ex);
+      throw new RecorderException("Invalid furniture library file " + furnitureLibraryLocation, ex);
     } catch (MissingResourceException ex) {
-      throw new RecorderException("Invalid furniture library file " + furnitureLibraryName, ex);
+      throw new RecorderException("Invalid furniture library file " + furnitureLibraryLocation, ex);
     }
   }
 
@@ -426,6 +451,10 @@ public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
       if (piece.isDoorOrWindow()) {
         // Write properties specific to doors and windows
         DoorOrWindow doorOrWindow = (DoorOrWindow)piece;
+        if (doorOrWindow.getCutOutShape() != null) {
+          writeProperty(writer, DefaultFurnitureCatalog.PropertyKey.DOOR_OR_WINDOW_CUT_OUT_SHAPE, i, 
+              doorOrWindow.getCutOutShape());
+        }
         if (doorOrWindow.getWallThickness() != 1) {
           writeProperty(writer, DefaultFurnitureCatalog.PropertyKey.DOOR_OR_WINDOW_WALL_THICKNESS, i, 
               doorOrWindow.getWallThickness() * doorOrWindow.getDepth());
@@ -449,9 +478,9 @@ public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
               sashStartAngle += " ";
               sashEndAngle += " ";
             }
-            sashXAxis += sashes [sashIndex].getXAxis() * doorOrWindow.getWidth();
-            sashYAxis += sashes [sashIndex].getYAxis() * doorOrWindow.getDepth();
-            sashWidth += sashes [sashIndex].getWidth() * doorOrWindow.getWidth();
+            sashXAxis += DECIMAL_FORMAT.format(sashes [sashIndex].getXAxis() * doorOrWindow.getWidth());
+            sashYAxis += DECIMAL_FORMAT.format(sashes [sashIndex].getYAxis() * doorOrWindow.getDepth());
+            sashWidth += DECIMAL_FORMAT.format(sashes [sashIndex].getWidth() * doorOrWindow.getWidth());
             sashStartAngle += Math.round(Math.toDegrees(sashes [sashIndex].getStartAngle()) * 100) / 100;
             sashEndAngle += Math.round(Math.toDegrees(sashes [sashIndex].getEndAngle()) * 100) / 100;
           }          
@@ -482,15 +511,15 @@ public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
                 lightSourceDiameter += " ";
               }
             }
-            lightSourceX += lightSources [lightIndex].getX() * light.getWidth();
-            lightSourceY += lightSources [lightIndex].getY() * light.getDepth();
-            lightSourceZ += lightSources [lightIndex].getZ() * light.getHeight();
+            lightSourceX += DECIMAL_FORMAT.format(lightSources [lightIndex].getX() * light.getWidth());
+            lightSourceY += DECIMAL_FORMAT.format(lightSources [lightIndex].getY() * light.getDepth());
+            lightSourceZ += DECIMAL_FORMAT.format(lightSources [lightIndex].getZ() * light.getHeight());
             lightSourceColor += "#" + Integer.toHexString(lightSources [lightIndex].getColor());
             if (lightSources [lightIndex].getDiameter() != null) {
               if (lightSourceDiameter == null) {
                 lightSourceDiameter = "";
               }
-              lightSourceDiameter += lightSources [lightIndex].getDiameter() * light.getWidth();
+              lightSourceDiameter += DECIMAL_FORMAT.format(lightSources [lightIndex].getDiameter() * light.getWidth());
             }
           }          
           writeProperty(writer, DefaultFurnitureCatalog.PropertyKey.LIGHT_SOURCE_X, i, lightSourceX);
@@ -715,6 +744,8 @@ public class FurnitureLibraryFileRecorder implements FurnitureLibraryRecorder {
       if (value.getClass().isArray()) {
         s = Arrays.toString((Object [])value);
         s = s.substring(1, s.length() - 1); // Remove brackets
+      } else if (value instanceof Float) {
+        s = DECIMAL_FORMAT.format(value);
       } else {
         s = value.toString();
       }
